package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.IntersectionTypeBinding18;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.Substitution;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;

public class Substitutor {
  public ReferenceBinding[] substitute(Substitution substitution, ReferenceBinding[] originalTypes) {
    if (originalTypes == null)
      return null; 
    ReferenceBinding[] substitutedTypes = originalTypes;
    for (int i = 0, length = originalTypes.length; i < length; i++) {
      ReferenceBinding originalType = originalTypes[i];
      TypeBinding substitutedType = substitute(substitution, (TypeBinding)originalType);
      if (!(substitutedType instanceof ReferenceBinding))
        return null; 
      if (substitutedType != originalType) {
        if (substitutedTypes == originalTypes)
          System.arraycopy(originalTypes, 0, substitutedTypes = new ReferenceBinding[length], 0, i); 
        substitutedTypes[i] = (ReferenceBinding)substitutedType;
      } else if (substitutedTypes != originalTypes) {
        substitutedTypes[i] = originalType;
      } 
    } 
    return substitutedTypes;
  }
  
  public TypeBinding substitute(Substitution substitution, TypeBinding originalType) {
    ParameterizedTypeBinding originalParameterizedType;
    ReferenceBinding originalEnclosing;
    ReferenceBinding substitutedEnclosing;
    TypeBinding[] originalArguments;
    TypeVariableBinding[] arrayOfTypeVariableBinding;
    TypeBinding[] substitutedArguments;
    ArrayBinding originalArrayType;
    TypeBinding originalLeafComponentType;
    TypeBinding substitute;
    WildcardBinding wildcard;
    IntersectionTypeBinding18 intersection;
    ReferenceBinding[] types;
    ReferenceBinding[] arrayOfReferenceBinding1;
    ReferenceBinding[] refSubsts;
    ReferenceBinding originalReferenceType;
    if (originalType == null)
      return null; 
    switch (originalType.kind()) {
      case 4100:
        return substitution.substitute((TypeVariableBinding)originalType);
      case 260:
        originalParameterizedType = (ParameterizedTypeBinding)originalType;
        originalEnclosing = originalType.enclosingType();
        substitutedEnclosing = originalEnclosing;
        if (originalEnclosing != null) {
          substitutedEnclosing = (ReferenceBinding)substitute(substitution, (TypeBinding)originalEnclosing);
          if (isMemberTypeOfRaw(originalType, substitutedEnclosing))
            return (TypeBinding)originalParameterizedType.environment.createRawType(
                originalParameterizedType.genericType(), substitutedEnclosing, originalType.getTypeAnnotations()); 
        } 
        originalArguments = originalParameterizedType.arguments;
        substitutedArguments = originalArguments;
        if (originalArguments != null) {
          if (substitution.isRawSubstitution())
            return (TypeBinding)originalParameterizedType.environment.createRawType(originalParameterizedType.genericType(), substitutedEnclosing, originalType.getTypeAnnotations()); 
          substitutedArguments = substitute(substitution, originalArguments);
        } 
        if (substitutedArguments != originalArguments || substitutedEnclosing != originalEnclosing)
          return (TypeBinding)originalParameterizedType.environment.createParameterizedType(
              originalParameterizedType.genericType(), substitutedArguments, substitutedEnclosing, originalType.getTypeAnnotations()); 
        break;
      case 68:
        originalArrayType = (ArrayBinding)originalType;
        originalLeafComponentType = originalArrayType.leafComponentType;
        substitute = substitute(substitution, originalLeafComponentType);
        if (substitute != originalLeafComponentType)
          return (TypeBinding)originalArrayType.environment.createArrayType(substitute.leafComponentType(), substitute.dimensions() + originalType.dimensions(), originalType.getTypeAnnotations()); 
        break;
      case 516:
      case 8196:
        wildcard = (WildcardBinding)originalType;
        if (wildcard.boundKind != 0) {
          TypeBinding originalBound = wildcard.bound;
          TypeBinding substitutedBound = substitute(substitution, originalBound);
          TypeBinding[] originalOtherBounds = wildcard.otherBounds;
          TypeBinding[] substitutedOtherBounds = substitute(substitution, originalOtherBounds);
          if (substitutedBound != originalBound || originalOtherBounds != substitutedOtherBounds) {
            if (originalOtherBounds != null) {
              TypeBinding[] bounds = new TypeBinding[1 + substitutedOtherBounds.length];
              bounds[0] = substitutedBound;
              System.arraycopy(substitutedOtherBounds, 0, bounds, 1, substitutedOtherBounds.length);
              TypeBinding[] glb = Scope.greaterLowerBound(bounds, null, substitution.environment());
              if (glb != null && glb != bounds) {
                substitutedBound = glb[0];
                if (glb.length == 1) {
                  substitutedOtherBounds = null;
                } else {
                  System.arraycopy(glb, 1, substitutedOtherBounds = new TypeBinding[glb.length - 1], 0, glb.length - 1);
                } 
              } 
            } 
            return (TypeBinding)wildcard.environment.createWildcard(wildcard.genericType, wildcard.rank, substitutedBound, substitutedOtherBounds, wildcard.boundKind, wildcard.getTypeAnnotations());
          } 
        } 
        break;
      case 32772:
        intersection = (IntersectionTypeBinding18)originalType;
        types = intersection.getIntersectingTypes();
        arrayOfReferenceBinding1 = substitute(substitution, types);
        refSubsts = new ReferenceBinding[arrayOfReferenceBinding1.length];
        System.arraycopy(arrayOfReferenceBinding1, 0, refSubsts, 0, arrayOfReferenceBinding1.length);
        return substitution.environment().createIntersectionType18(refSubsts);
      case 4:
        if (!originalType.isMemberType())
          break; 
        originalReferenceType = (ReferenceBinding)originalType.unannotated();
        originalEnclosing = originalType.enclosingType();
        substitutedEnclosing = originalEnclosing;
        if (originalEnclosing != null) {
          substitutedEnclosing = (ReferenceBinding)substitute(substitution, (TypeBinding)originalEnclosing);
          if (isMemberTypeOfRaw(originalType, substitutedEnclosing))
            return (TypeBinding)substitution.environment().createRawType(originalReferenceType, substitutedEnclosing, originalType.getTypeAnnotations()); 
        } 
        if (substitutedEnclosing != originalEnclosing)
          return substitution.isRawSubstitution() ? 
            (TypeBinding)substitution.environment().createRawType(originalReferenceType, substitutedEnclosing, originalType.getTypeAnnotations()) : 
            (TypeBinding)substitution.environment().createParameterizedType(originalReferenceType, null, substitutedEnclosing, originalType.getTypeAnnotations()); 
        break;
      case 2052:
        originalReferenceType = (ReferenceBinding)originalType.unannotated();
        originalEnclosing = originalType.enclosingType();
        substitutedEnclosing = originalEnclosing;
        if (originalEnclosing != null) {
          substitutedEnclosing = originalType.isStatic() ? (ReferenceBinding)substitution.environment().convertToRawType((TypeBinding)originalEnclosing, true) : 
            (ReferenceBinding)substitute(substitution, (TypeBinding)originalEnclosing);
          if (isMemberTypeOfRaw(originalType, substitutedEnclosing))
            return (TypeBinding)substitution.environment().createRawType(originalReferenceType, substitutedEnclosing, originalType.getTypeAnnotations()); 
        } 
        if (substitution.isRawSubstitution())
          return (TypeBinding)substitution.environment().createRawType(originalReferenceType, substitutedEnclosing, originalType.getTypeAnnotations()); 
        arrayOfTypeVariableBinding = originalReferenceType.typeVariables();
        substitutedArguments = substitute(substitution, (TypeBinding[])arrayOfTypeVariableBinding);
        return (TypeBinding)substitution.environment().createParameterizedType(originalReferenceType, substitutedArguments, substitutedEnclosing, originalType.getTypeAnnotations());
    } 
    return originalType;
  }
  
  private static boolean isMemberTypeOfRaw(TypeBinding originalType, ReferenceBinding substitutedEnclosing) {
    return (substitutedEnclosing != null && substitutedEnclosing.isRawType() && 
      originalType instanceof ReferenceBinding && !((ReferenceBinding)originalType).isStatic());
  }
  
  public TypeBinding[] substitute(Substitution substitution, TypeBinding[] originalTypes) {
    if (originalTypes == null)
      return null; 
    TypeBinding[] substitutedTypes = originalTypes;
    for (int i = 0, length = originalTypes.length; i < length; i++) {
      TypeBinding originalType = originalTypes[i];
      TypeBinding substitutedParameter = substitute(substitution, originalType);
      if (substitutedParameter != originalType) {
        if (substitutedTypes == originalTypes)
          System.arraycopy(originalTypes, 0, substitutedTypes = new TypeBinding[length], 0, i); 
        substitutedTypes[i] = substitutedParameter;
      } else if (substitutedTypes != originalTypes) {
        substitutedTypes[i] = originalType;
      } 
    } 
    return substitutedTypes;
  }
}
